#!/usr/bin/env python
# encoding: utf-8

import os
import time
import platform

from Naked.toolshed.file import FileReader, FileWriter


class DoxxCache(object):
    def __init__(self):
        self.system = platform.system()
        self.package_repo_list_file = "list.txt"        # the list of packages in the package repository
        self.package_repo_json_file = "packages.json"   # the name:description JSON file of packages in the package repository
    
    #################################
    #
    #  Cache Writer Methods
    #
    #################################
    def cache_textfile(self, file_name, file_text):
        cache_dir_path = self._get_platform_specific_cache_dirpath()    # get the platform specific path to the cache directory
        
        # confirm that the local variable was assigned based upon user system type
        if cache_dir_path is not None:
            if not os.path.isdir(cache_dir_path):
                os.mkdir(cache_dir_path)
            cache_file_path = os.path.join(cache_dir_path, file_name)    # create the path to the file in the cache directory
            if self._write_text_file(cache_file_path, file_text):  # write file to cache storage
                return True   # return True if cached file write OK
            else:
                return False  # return False if fails to write
        else:
            return False  # return False if unable to detect user system

    def cache_packagerepo_list(self, file_text):
        if self.cache_textfile(self.package_repo_list_file, file_text):
            return True
        else:
            return False
        
    def cache_packagerepo_json(self, json_string):
        if self.cache_textfile(self.package_repo_json_file, json_string):
            return True
        else:
            return False
    
    #################################
    #
    #  Cache Reader Methods
    #
    #################################
    def get_cached_textfile(self, file_name):
        cache_file_dir = self._get_platform_specific_cache_dirpath()
        cache_file_path = os.path.join(cache_file_dir, file_name)
        return self._read_text_file(cache_file_path)
    
    def get_cached_packagerepo_list(self):
        cache_file_dir = self._get_platform_specific_cache_dirpath()
        cache_file_path = os.path.join(cache_file_dir, self.package_repo_list_file)
        return self._read_text_file(cache_file_path)
    
    def get_cached_packagerepo_json(self):
        cache_file_dir = self._get_platform_specific_cache_dirpath()
        cache_file_path = os.path.join(cache_file_dir, self.package_repo_json_file)
        return self._read_text_file(cache_file_path)        
    
    #################################
    #
    #  Utility Methods
    #
    #################################
    
    #################################
    # PUBLIC
    #################################
    
    def cached_file_exists(self, file_name):
        cache_file_dir = self._get_platform_specific_cache_dirpath()
        cache_file_path = os.path.join(cache_file_dir, file_name)
        if os.path.isfile(cache_file_path):
            return True
        else:
            return False

    def does_cache_file_require_update(self, file_name, diff_seconds_for_update):
        # diff_seconds_for_update is the number of seconds to consider stale data and need for update
        cache_file_dir = self._get_platform_specific_cache_dirpath()
        cache_file_path = os.path.join(cache_file_dir, file_name)
        current_time = time.time()
        if os.path.isfile(cache_file_path):
            if current_time - os.path.getmtime(cache_file_path) > diff_seconds_for_update:
                return True
            else:
                return False
        else:
            return True  # need to update if the file does not exist
    
    #################################
    # PRIVATE
    #################################
    
    def _get_mac_cachedir(self):
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, "Library", "doxx")    
    
    def _get_linux_cachedir(self):
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, '.doxx')
    
    def _get_windows_cachedir(self):
        return os.path.join(os.environ['APPDATA'], 'doxx')
    
    def _write_text_file(self, file_path, file_text):
        try:
            fw = FileWriter(file_path)
            fw.write(file_text)
            return True
        except Exception:
            return False
        
    def _read_text_file(self, file_path):
        try:
            fr = FileReader(file_path)
            return fr.read()  # return the file text
        except Exception:
            return ""  # return empty string if there is an exception during the read
        
    def _get_platform_specific_cache_dirpath(self):
        # detect user system
        if self.system == "Darwin":
            return self._get_mac_cachedir()
        elif self.system == "Linux":
            return self._get_linux_cachedir()
        elif self.system == "Windows":
            return self._get_windows_cachedir()
        else:
            return None
    
    


#!/usr/bin/env python
# encoding: utf-8


class OfficialPackage(object):
    def __init__(self):
        self.binary_package_url_prefix = "https://github.com/doxx-repo/"
        self.binary_package_url_postfix = "/releases/download/current/"
        self.package_key_file_url_prefix = "https://github.com/doxx-repo/"
        self.package_key_file_url_postfix = "/releases/download/current/pkey.yaml"
        
        self.master_package_list_text_url = "https://doxx-cdn.s3.amazonaws.com/list.txt"
        self.master_package_description_json_url = "https://doxx-cdn.s3.amazonaws.com/packages.json"
        
        # Release file URL's
        # https://github.com/doxx-repo/{package-name}/releases/download/current/{package-name}.tar.gz
        # https://github.com/doxx-repo/{package-name}/releases/download/current/{package-name}.zip
        # https://github.com/doxx-repo/{package-name}/releases/download/current/pkey.yaml
    
    # Attribute getter methods
    
    def get_package_targz_url(self, package_name):
        normalized_package_name = package_name.lower().strip()
        return self.binary_package_url_prefix + normalized_package_name + self.binary_package_url_postfix + normalized_package_name + '.tar.gz'
    
    def get_package_zip_url(self, package_name):
        normalized_package_name = package_name.lower().strip()
        return self.binary_package_url_prefix + normalized_package_name + self.binary_package_url_postfix + normalized_package_name + '.zip'     
    
    def get_package_key_url(self, package_name):
        normalized_package_name = package_name.lower().strip()
        return self.package_key_file_url_prefix + normalized_package_name + self.package_key_file_url_postfix
    
    def get_master_package_list_url(self):
        return self.master_package_list_text_url
    
    def get_master_package_description_json_url(self):
        return self.master_package_description_json_url
    

    
    


#!/usr/bin/env python
# encoding: utf-8

from Naked.toolshed.file import FileWriter
from Naked.toolshed.network import HTTP
from Naked.toolshed.system import stdout, stderr
from doxx.datatypes.package import OfficialPackage


def run_pullkey(package_name):
    normalized_package_name = package_name.lower().strip()
    package = OfficialPackage()
    key_file_url = package.get_package_key_url(normalized_package_name)
    try:
        stdout("[*] doxx: Pulling remote key file...")
        http = HTTP(key_file_url)
        if http.get_status_ok():
            key_file_text = http.res.text
            fr = FileWriter('key.yaml')
            try:
                fr.write(key_file_text)
            except Exception as e:
                stderr("[!] doxx: Unable to write the 'key.yaml' file to disk. Error: " + str(e), exit=1)
            stdout("[*] doxx: Key file pull complete")
        elif http.res.status_code == 404:
            stderr("[!] doxx: Unable to pull the key file because the requested package could not be found. (HTTP status code: 404)", exit=1)
        else:
            stderr("[!] doxx: Unable to pull the key file.  (HTTP status code: " + str(http.res.status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull the key file. Error: " + str(e), exit=1)
        


#!/usr/bin/env python
# encoding: utf-8

#------------------------------------------------------------------------------
# doxx
# Copyright 2015 Christopher Simpkins
# MIT license
#------------------------------------------------------------------------------

# Application start
def main():
    import sys
    from os.path import basename
    from Naked.commandline import Command
    from Naked.toolshed.system import stdout, stderr, is_dir, is_file, cwd
    from doxx.commands.build import Builder
    from doxx.commands.make import Maker

    #------------------------------------------------------------------------------------------
    # [ Instantiate command line object ]
    #   used for all subsequent conditional logic in the CLI application
    #------------------------------------------------------------------------------------------
    c = Command(sys.argv[0], sys.argv[1:])
    #------------------------------------------------------------------------------------------
    # [ Command Suite Validation ] - early validation of appropriate command syntax
    # Test that user entered at least one argument to the executable, print usage if not
    #------------------------------------------------------------------------------------------
    if not c.command_suite_validates():
        from doxx.settings import usage as doxx_usage
        print(doxx_usage)
        sys.exit(1)
    #------------------------------------------------------------------------------------------
    # [ NAKED FRAMEWORK COMMANDS ]
    # Naked framework provides default help, usage, and version commands for all applications
    #   --> settings for user messages are assigned in the lib/doxx/settings.py file
    #------------------------------------------------------------------------------------------
    if c.help():      # User requested doxx help information
        from doxx.settings import help as doxx_help
        print(doxx_help)
        sys.exit(0)
    elif c.usage():   # User requested doxx usage information
        from doxx.settings import usage as doxx_usage
        print(doxx_usage)
        sys.exit(0)
    elif c.version():   # User requested doxx version information
        from doxx.settings import app_name, major_version, minor_version, patch_version
        version_display_string = app_name + ' ' + major_version + '.' + minor_version + '.' + patch_version
        print(version_display_string)
        sys.exit(0)
    #------------------------------------------------------------------------------------------
    # [ PRIMARY COMMAND LOGIC ]
    #   Enter your command line parsing logic below
    #------------------------------------------------------------------------------------------
    elif c.cmd == "build":
        if c.argc > 1:
            key_path = c.arg1
        else:
            key_path = "key.yaml"
        stdout("[*] doxx: Build started with the key file '" + key_path + "'...")
        b = Builder(key_path)
        b.run()
        stdout("[*] doxx: Build complete.")
    elif c.cmd == "clean":
        from doxx.commands.clean import run_clean
        run_clean()  # execute the clean routines
    elif c.cmd == "make":
        if c.argc > 1:
            if c.cmd2 == "key":  # secondary command        
                m = Maker()
                if c.argc > 2:
                    m.make_key(c.arglp)
                else:
                    m.make_key("key.yaml")
            elif c.cmd2 == "template":  # secondary command
                m = Maker()
                if c.argc > 2:
                    m.make_template(c.arglp)
                else:
                    m.make_template("stub.doxt")  # default name is 'stub.doxt' for new template if not specified by user
            elif c.cmd2 == "project":
                m = Maker()
                m.make_project()
            else:
                stderr("Usage: doxx make [key|project|template]", exit=1)
        else:
            stderr("[!] doxx: Please include the secondary command 'key', 'project', or 'template' with the 'make' command.", exit=1)
    elif c.cmd == "pack":
        from doxx.commands.pack import tar_gzip_package_directory, zip_package_directory 
        if c.argc > 1:
            if c.cmd2 == "zip":
                if c.argc > 2:  # request for zip with a directory path
                    if is_dir(c.arglp):
                        zip_package_directory(c.arglp, c.arglp)
                    else:
                        stderr("[!] doxx: '" + c.arglp + "' does not appear to be a directory.  Please enter the path to your project directory.", exit=1)
                else:  # request for zip with current working directory
                    stderr("[!] doxx: Please include your project directory as an argument to the zip command", exit=1)
            else:  # request for tar.gz with a directory path
                if is_dir(c.arglp):
                    tar_gzip_package_directory(c.arglp, c.arglp)
                else:
                    stderr("[!] doxx: '" + c.arglp + "' does not appear to be a directory.  Please enter the path to your project directory.", exit=1)
        else:  # request for tar.gz in current working directory
            root_dir = cwd()
            archive_name = basename(root_dir)
            tar_gzip_package_directory(archive_name, root_dir)
        # end of the pack command
        stdout("[*] doxx: Pack complete")
    elif c.cmd == "pull":
        if c.argc > 1:
            from doxx.commands.pull import run_pull
            run_pull(c.arg1)
            stdout("[*] doxx: Pull complete")
        else:
            stderr("[!] doxx: Please include the URL for the archive that you would like to pull.", exit=1)
    elif c.cmd == "pullkey":
        if c.argc > 1:
            from doxx.commands.pullkey import run_pullkey
            run_pullkey(c.arg1)
        else:
            stderr("[!] doxx: Please include a package name with the pullkey command", exit=1)
    elif c.cmd == "search":
        if c.argc > 1:
            from doxx.commands.search import run_search
            run_search(c.arg1)
        else:
            stderr("[!] doxx: Please include a search string after your command.", exit=1)
    elif c.cmd == "unpack":
        if c.argc > 1:
            if is_file(c.arg1):
                from doxx.commands.unpack import unpack_run, remove_compressed_archive_file
                unpack_run(c.arg1)
                remove_compressed_archive_file(c.arg1)
                stdout("[*] doxx: Unpack complete")
            else:
                stderr("[!] doxx: '" + c.arg1 + "' does not appear to be a file.  Please include a path to your compressed file.", exit=1)
        else:
            stderr("[!] doxx: Please include a path to your file.", exit=1)
    elif c.cmd == "whatis":
        if c.argc > 1:
            from doxx.commands.whatis import run_whatis
            run_whatis(c.arg1)
        else:
            stderr("[!] doxx: Please enter a package name following your command.", exit=1)
    #------------------------------------------------------------------------------------------
    # UNDOCUMENTED TESTING COMMANDS
    #  
    #------------------------------------------------------------------------------------------    
    elif c.cmd == "repoupdate":
        from doxx.commands.repoupdate import run_repoupdate
        run_repoupdate()
    
    #------------------------------------------------------------------------------------------
    # [ DEFAULT MESSAGE FOR MATCH FAILURE ]
    #  Message to provide to the user when all above conditional logic fails to meet a true condition
    #------------------------------------------------------------------------------------------
    else:
        stderr("[!] doxx: Could not complete the command that you entered.  Please try again.")
        sys.exit(1) #exit

if __name__ == '__main__':
    main()
    
    


#!/usr/bin/env python
# encoding: utf-8

#------------------------------------------------------------------------------
# Application Name
#------------------------------------------------------------------------------
app_name = 'doxx'

#------------------------------------------------------------------------------
# Version Number
#------------------------------------------------------------------------------
major_version = "0"
minor_version = "9"
patch_version = "0"

#------------------------------------------------------------------------------
# Debug Flag (switch to False for production release code)
#------------------------------------------------------------------------------
debug = False

#------------------------------------------------------------------------------
# Usage String
#------------------------------------------------------------------------------
usage = """
doxx <option> [command] <argument>

Use 'doxx help' for additional information
"""

#------------------------------------------------------------------------------
# Help String
#------------------------------------------------------------------------------
help = """
----------------------------------------------------------
 doxx
 A simple, flexible text file project build system
 Copyright 2015 Christopher Simpkins
 MIT license
 Source: http://github.com/chrissimpkins/doxx
   Docs: http://doxx.org
----------------------------------------------------------

USAGE
  doxx <option> [command] <argument>
  
GENERAL COMMANDS
  build    render string replacements in template files and build project
  clean    remove doxx project files from a project directory
  make     generate key, template, or project file stubs
  pack     create a tar.gz or zip archive file for project distribution
  pull     pull a key, template, or project archive from a remote repository
  unpack   unpack a tar.gz or zip project archive

PACKAGE REPOSITORY COMMANDS
  pull     pull a project package from the doxx Package Repository
  pullkey  pull the key file from a doxx Package Repository project
  search   search the doxx Package Repository by keyword or project name
  whatis   get descriptions of Package Repository packages by project name

OPTIONS
  -h | --help      view application help
       --usage     view application usage
  -v | --version   view installed application version
"""




#!/usr/bin/env python
# encoding: utf-8

from os import remove
from shutil import rmtree
from Naked.toolshed.system import list_filter_files, list_filter_files_cwd, list_all_files, dir_exists, file_exists, make_path, stderr, stdout


def run_clean():
    _remove_key()
    _remove_doxt()
    stdout("[*] doxx: Clean complete.")


def _remove_key():
    # remove key.yaml
    if file_exists('key.yaml'):
        try:
            remove('key.yaml')
            stdout("[-] doxx: -- key.yaml ... removed")
        except Exception as e:
            stderr("[!] doxx: Unable to remove 'key.yaml'. Error: " + str(e), exit=0)
    
    # remove pkey.yaml (project archives)
    if file_exists('pkey.yaml'):
        try:
            remove('pkey.yaml')
            stdout("[-] doxx: -- pkey.yaml ... removed")
        except Exception as e:
            stderr("[!] doxx: Unable to remove 'pkey.yaml'. Error: " + str(e), exit=0)
    
    # remove project.yaml (project archives)
    if file_exists('project.yaml'):
        try:
            remove('project.yaml')
            stdout("[-] doxx: -- project.yaml ... removed")
        except Exception as e:
            stderr("[!] doxx: Unable to remove 'project.yaml'. Error: " + str(e), exit=0)     

     
def _remove_doxt():
    # check main directory
    cwd_doxt_list = list_filter_files_cwd(".doxt")
    
    # check for a templates directory
    if dir_exists('templates'):
        templates_dir_doxt_list = list_filter_files(".doxt", "templates")
    else:
        templates_dir_doxt_list = []
    
    # remove template files from the working directory
    if len(cwd_doxt_list) > 0:
        for doxt_file in cwd_doxt_list:
            try:
                remove(doxt_file)
                stdout("[-] doxx: -- " + doxt_file + " ... removed")
            except Exception as e:
                stderr("[!] doxx: Unable to remove the file " + doxt_file + "'. Error: " + str(e), exit=0)
    
    # remove any template files from the templates directory
    if len(templates_dir_doxt_list) > 0:
        for doxt_file in templates_dir_doxt_list:
            new_doxt_path = make_path('templates', doxt_file)
            remove(new_doxt_path)
            stdout("[-] doxx: -- " + new_doxt_path + " ... removed")
            
    # if there are files still remaining in the templates directory, leave it
    # otherwise, remove the templates directory as well
    if dir_exists('templates'):
        remaining_template_file_list = list_all_files('templates')  # get the remaining non-.doxt files in the directory
        if len(remaining_template_file_list) > 0:
            pass  # do nothing, skip the removal of the 'templates' directory from the project because other files are present in it
        else:
            try:
                rmtree('templates')
                stdout("[-] doxx: -- templates (directory) ... removed")
            except Exception as e:
                stderr("[!] doxx: Unable to remove the 'templates' directory.  Error: " + str(e), exit=1)
                
                


#!/usr/bin/env python
# encoding: utf-8

import difflib


class FuzzySearcher(object):
    def __init__(self, needle):
        self.needle = needle
        self.needle_normalized = needle.lower().strip()
        self.needle_alpha = self._get_alphabetical_string(self.needle_normalized)
        self.needle_word_count = len(needle.split(" "))
        self.needle_length = len(self.needle_normalized)
    
    # Search Types
    
    # test match to the entire haystack string
    def full_string_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        return difflib.SequenceMatcher(None, self.needle_normalized, normalized_haystack).ratio()
    
    # partial match ratio for first index item after split on '-' character
    def partial_firstindexitem_dashsplit_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        haystack_tokens = normalized_haystack.split('-')
        first_haystack_token = haystack_tokens[0]
        return difflib.SequenceMatcher(None, self.needle_normalized, first_haystack_token).ratio()
    
    # partial match ratio for needle length slice from the haystack
    def partial_startslice_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        needle_length = len(self.needle_normalized)
        if len(normalized_haystack) > needle_length:
            sliced_haystack = normalized_haystack[0:needle_length]
            return difflib.SequenceMatcher(None, self.needle_normalized, sliced_haystack).ratio()
            
        else:
            return difflib.SequenceMatcher(None, self.needle_normalized, normalized_haystack).ratio()
        
    # split haystack into tokens on '-' character delimiters, return best ratio from tokens (use for needle word length = 1)
    def partial_dashsplit_tokens_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        haystack_tokens = normalized_haystack.split("-")
        best_ratio = 0
        for token in haystack_tokens:
            the_ratio = difflib.SequenceMatcher(None, self.needle_normalized, token).ratio()
            if the_ratio > best_ratio:
                best_ratio = the_ratio
        return best_ratio
    
    # attempt match over same number of word tokens as are present in the needle, sorted in alpha order (for multi-word searches)
    def partial_nword_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        haystack_token_list = normalized_haystack.split('-')
        haystack_word_count = len(haystack_token_list)
        if haystack_word_count >= self.needle_word_count:
            first = 0
            last = self.needle_word_count
            best_ratio = 0
            for x in range(haystack_word_count - (self.needle_word_count - 1)):
                if last <= haystack_word_count:
                    sub_haystack_list = haystack_token_list[first:last]
                    sorted_sub_haystack_list = sorted(sub_haystack_list)
                    sorted_sub_haystack_string = " ".join(sorted_sub_haystack_list)
                    token_match_ratio = difflib.SequenceMatcher(None, self.needle_alpha, sorted_sub_haystack_string).ratio()
                    if token_match_ratio > best_ratio:
                        best_ratio = token_match_ratio
                    
                    first += 1  # iterate the positions of the test string slice
                    last += 1
            return best_ratio
        else:
            return difflib.SequenceMatcher(None, self.needle_normalized, normalized_haystack).ratio()
        
    # set intersection between needle and haystack tokens with addition of remaining parts of strings (use > 0.7 as threshold)
    def partial_set_ratio(self, haystack):
        normalized_haystack = haystack.lower().strip()
        haystack_alpha = sorted(normalized_haystack.split('-'))
        haystack_set = set(haystack_alpha)
        needle_set = set(self.needle_alpha.split(" "))
        
        intersection_set = sorted(needle_set.intersection(haystack_set))
        needle_difference = sorted(needle_set.difference(haystack_set))
        haystack_difference = sorted(haystack_set.difference(needle_set))
        
        if len(intersection_set) > 0:
            string_one = " ".join(intersection_set)
            
            if len(needle_difference) > 0:
                string_two = string_one + " " + " ".join(needle_difference)
            else:
                string_two = string_one  # if there were no tokens in difference, it is just the intersection
            
            if len(haystack_difference) > 0:
                string_three = string_one + " " + " ".join(haystack_difference)
            else:
                string_three = string_one  # if there were no tokens in the difference, it is just the intersection
            
            token_match_ratio_one = difflib.SequenceMatcher(None, string_one, string_two).ratio()
            token_match_ratio_two = difflib.SequenceMatcher(None, string_one, string_three).ratio()
            token_match_ratio_three = difflib.SequenceMatcher(None, string_two, string_three).ratio()
            
            # return an evenly weighted average of the match ratios
            weighted_average_ratio = (0.333 * token_match_ratio_one) + (0.333 * token_match_ratio_two) + (0.333 * token_match_ratio_three)
            return weighted_average_ratio
        else:
            return 0  # if there is no intersection tokens between the needle and haystack tokens, return 0
    
    # Utilities
    
    def _get_alphabetical_string(self, pre_string):
        if " " in pre_string:
            alpha_list = sorted(pre_string.split(" "))
            alpha_string = " ".join(alpha_list)
            return alpha_string
        else:
            return pre_string
        
    
    


#!/usr/bin/env python
# encoding: utf-8

import sys
import os
from multiprocessing import Process, Lock

from Naked.toolshed.file import FileWriter
from Naked.toolshed.system import cwd, dir_exists, file_exists, make_dirs, make_path, stderr, stdout
from Naked.toolshed.python import is_py2

from doxx.datatypes.template import DoxxTemplate, RemoteDoxxTemplate
from doxx.datatypes.key import DoxxKey
from doxx.commands.pull import run_pull, is_url
from doxx.commands.unpack import unpack_run

# need a different template for Python 2 & 3
if is_py2():    
    from doxx.renderer.inkpy2 import Template as InkTemplate
    from doxx.renderer.inkpy2 import Renderer as InkRenderer
else:
    from doxx.renderer.inkpy3 import Template as InkTemplate
    from doxx.renderer.inkpy3 import Renderer as InkRenderer


def multi_process_build(key, key_path):
    processes = []       # list of spawned processes
    iolock = Lock()      # file read / write lock
    outputlock = Lock()  # stdout / stderr writes lock
    
    # ## SINGLE PROCESS
    # for template in key.meta_data['templates']:
        # b = Builder()
        # b.set_key_data(key)
        # b.single_template_run(template)
    
    ## MULTI-PROCESS    
    # create worker processes
    for template in key.meta_data['templates']:
        p = Process(target=single_process_runner, args=(template, key, key_path, iolock, outputlock))
        p.start()
        processes.append(p)
    # join worker processes upon completion or with 60 second timeout
    for process in processes:
        process.join(timeout=60)
    
    # ## zombie process finder for testing
    # active_child_processes = active_children()
    
    # if len(active_child_processes) > 0:
        # print("active children detected")
    # else:
        # print("no active children detected")


def single_process_runner(template_path, key, key_path, iolock, outputlock):
    b = Builder(key_path)
    b.set_key_data(key)
    b.multi_process_run(template_path, iolock, outputlock)


class Builder(object):
    """The Builder class renders doxx templates from user provided keys"""
    def __init__(self, key_path): 
        self.key_path = key_path   # the local key file path: key.yaml or as specified by user
        self.key_data = {}
        self.no_key_replacements = False
    
    def run(self):
        doxxkey = DoxxKey(self.key_path)
        # detect single vs multiple keys in the template and execute replacements with every requested template
        self.set_key_data(doxxkey)  # assign key data from the doxx Key
        
        try:
            if doxxkey.project_key is True:  # the key is set to run on a local or remote project archive
                self.project_archive_run(doxxkey)
            elif doxxkey.multi_template_key is True:  # the key is set to run on multiple local or remote template files
                multi_process_build(doxxkey, self.key_path)
            else:
                self.single_template_run(doxxkey.meta_data['template']) # the key is set to run on a single local or remote template file
        except Exception as e:
            stderr("[!] doxx: Error: " + str(e), exit=1)
            
    
    def set_key_data(self, key):
        """used to set the key data"""
        self.key_data = key.key_data
        self.no_key_replacements = key.no_replacements  # defined True if there were no replacement keys in the key file, used to indicate skip replacements in the template(s)
        
    
    def single_template_run(self, template_path):
        """Render replacements using a single template file as defined in a doxx key file (public method)"""
        #----------------------------------------------------------------------------
        # NOTE : changes in this method require the same changes to multi_process_run
        #----------------------------------------------------------------------------       
        ## Load the data
        # remote templates
        if len(template_path) > 6 and (template_path[0:7] == "http://" or template_path[0:8] == "https://"):
            template = RemoteDoxxTemplate(template_path)
            try:
                result = template.load_data()
                if result[0] == False:  # if the method responds False, then HTTP data load did not work
                    stderr(result[1], exit=1)  # write out the returned error message in result[1] position of the tuple
                    # halt execution if unsuccessful
            except Exception as e:
                stderr("[!] doxx: Unable to load the remote template file '" + template_path + "'. Error message: " + str(e), exit=1)
        # local templates        
        elif file_exists(template_path):
            template = DoxxTemplate(template_path)
            try:
                template.load_data()
            except Exception as e:
                stderr("[!] doxx: Unable to read the local template file '" + template_path + "'. Error message: " + str(e), exit=1)
        else:
            stderr("[!] doxx: Unable to find the requested template file " + template_path, exit=1)  # print error message and halt execution of application
         
        ## Split the data  
        try:
            template.split_data()
        except Exception as e:
            stderr("[!] doxx: Unable to parse the template data.  Please verify the template syntax and try again.  Error message: " + str(e), exit=1)
        
        ## Parse data for errors
        error_parse_result = template.parse_template_for_errors()
        
        if error_parse_result[0] == True:          # if there was a parsing error
            stderr(error_parse_result[1], exit=1)  # print the returned error message to stderr and exit application
        
        ## Then parse the template text and load instance attributes for the text replacement with Ink below
        try:
            template.parse_template_text()
        except Exception as e:
            stderr("[!] doxx: An error occurred while parsing your template file. Error message: " + str(e), exit=1)
    
        # determine whether this is a verbatim template file (no replacements) or the key file did not include replacement keys
        if template.verbatim is True or self.no_key_replacements is True:
            # write template.text out verbatim
            try:
            # if the requested destination directory path does not exist, make it
                outfile_dir_path = make_path(os.path.dirname(self.key_path), os.path.dirname(template.outfile))
                if not outfile_dir_path == '' and not dir_exists(outfile_dir_path):
                    make_dirs(outfile_dir_path)
                # write the file
                outfile_path = make_path(os.path.dirname(self.key_path), template.outfile)
                fw = FileWriter(outfile_path)
                fw.write(template.text)
                stdout("[+] doxx: '" + outfile_path + "' build... check")
            except Exception as e:
                stderr("[!] doxx: There was a file write error. Error message: " + str(e), exit=1)        
        else:
            # template meta data is in template.meta_data
            # template text is in template.text
            # perform the text replacements:         
            try:
                ink_template = InkTemplate(template.text)
                ink_renderer = InkRenderer(ink_template, self.key_data)
                rendered_text = ink_renderer.render()
            except Exception as e:
                stderr("[!] doxx: An error occurred during the text replacement attempt.  Error message: " + str(e), exit=1)
        
            # if the requested destination directory path does not exist, make it
            outfile_dir_path = make_path(os.path.dirname(self.key_path), os.path.dirname(template.outfile))
            if not outfile_dir_path == '' and not dir_exists(outfile_dir_path):
                make_dirs(outfile_dir_path)
    
            # write rendered file to disk
            try:
                outfile_path = make_path(os.path.dirname(self.key_path), template.outfile)
                fw = FileWriter(outfile_path)
                fw.write(rendered_text)
                stdout("[+] doxx: -- " + outfile_path + " ... check")
            except Exception as e:
                stderr("[!] doxx: There was an error with the rendered file write. Error message: " + str(e), exit=1)
            
    def multi_process_run(self, template_path, iolock, outputlock):
        """Render replacements over multiple template files as defined in doxx key file using multiple processes (public method)"""
        #-------------------------------------------------------------------------------
        # NOTE : changes in this method require the same changes to single_template_run
        #-------------------------------------------------------------------------------
        ## Load the data
        if len(template_path) > 6 and (template_path[0:7] == "http://" or template_path[0:8] == "https://"):
            template = RemoteDoxxTemplate(template_path)
            try:
                result = template.load_data()  # load remote template file through HTTP or HTTPS protocol
                if result[0] == False:  # if the method responds False, then HTTP data load was not successful
                    outputlock.acquire()
                    stderr(result[1], exit=0)  # write out the returned error message in result[1] position of the tuple
                    outputlock.release()
                    sys.exit(1)  # release the lock before raising SystemExit
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: Unable to load the remote template file '" + template_path + "'. Error message: " + str(e), exit=0)
                outputlock.release()
                sys.exit(1)  # release the lock before raising SystemExit
        elif file_exists(template_path):
            template = DoxxTemplate(template_path)
            
            iolock.acquire()  # acquire the IO lock for template file read
            try:
                template.load_data()  # load local data
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: Unable to read the local template file '" + template_path + "'. Error message: " + str(e), exit=0)
                outputlock.release()
                sys.exit(1)  # release the lock before raising SystemExit
            iolock.release()  # release the IO lock for template file read
        else:
            outputlock.acquire()  # acquire the stderr lock
            stdout("[!] doxx: Unable to find the requested template file " + template_path)  # print error message in standard output, multi-file run so do not end execution of application       
            outputlock.release()  # release the stderr lock

        ## Split the data
        try:
            template.split_data()  # split the data sections
        except Exception as e:
            outputlock.acquire()
            stderr("[!] doxx: Unable to parse the template data.  Please verify the template syntax and try again.  Error message: " + str(e), exit=0)
            outputlock.release()
            sys.exit(1)  # release the lock before raising SystemExit
            
        ## Parse the data for errors
        error_parse_result = template.parse_template_for_errors()
        
        if error_parse_result[0] == True:          # if there was a parsing error
            outputlock.acquire()
            stderr(error_parse_result[1], exit=0)  # print the returned error message to stderr and exit application
            outputlock.release()
            sys.exit(1)  # release the lock before raising SystemExit
    
        ## Parse the template text
        try:
            template.parse_template_text()
        except Exception as e:
            outputlock.acquire()
            stderr("[!] doxx: An error occurred while parsing your template file. Error message: " + str(e), exit=0)
            outputlock.release()
            sys.exit(1)  # release the lock before raising SystemExit
        
        # determine whether this is a verbatim template file (no replacements) or the key file did not include replacement keys
        if template.verbatim is True or self.no_key_replacements is True:
            # write template.text out verbatim
            try:
                # if the requested destination directory path does not exist, make it
                outfile_dir_path = make_path(os.path.dirname(self.key_path), os.path.dirname(template.outfile))
                try:
                    iolock.acquire()
                    if not outfile_dir_path == '' and not dir_exists(outfile_dir_path):
                        make_dirs(outfile_dir_path)
                    iolock.release()
                except Exception as e:
                    iolock.release()  # release the lock then re-raise the exception
                    raise e
                # write the file
                outfile_path = make_path(os.path.dirname(self.key_path), template.outfile)
                # then write the file out verbatim
                try:
                    iolock.acquire()
                    fw = FileWriter(outfile_path)
                    fw.write(template.text)
                    iolock.release()
                except Exception as e:
                    iolock.release()  # catch and release the iolock before re-raising the exception
                    raise e
                
                outputlock.acquire()
                stdout("[+] doxx: -- " + outfile_path + " ... check")
                outputlock.release()
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: There was a file write error with '" + template_path + "'. Error message: " + str(e), exit=0)
                outputlock.release()
                sys.exit(1)  # release the lock before raising SystemExit
        else:
            # template meta data is in template.meta_data
            # template text is in template.text
            # perform the text replacements:
            try:
                ink_template = InkTemplate(template.text)
                ink_renderer = InkRenderer(ink_template, self.key_data)
                rendered_text = ink_renderer.render()
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: An error occurred during the text replacement attempt.  Error message: " + str(e), exit=0)            
                outputlock.release()
                sys.exit(1)  # release the lock before raising SystemExit
        
            # if the requested destination directory path does not exist, make it
            outfile_dir_path = make_path(os.path.dirname(self.key_path), os.path.dirname(template.outfile))
            try:
                iolock.acquire()
                if not outfile_dir_path == '' and not dir_exists(outfile_dir_path):
                    make_dirs(outfile_dir_path)
                iolock.release()
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: Unable to create directory path '" + outfile_dir_path + "' for your file write. Error: " + str(e), exit=0)
                outputlock.release()
                iolock.release()
                sys.exit(1)  # release the iolock before raising SystemExit
                
            outfile_path = make_path(os.path.dirname(self.key_path), template.outfile)
            
            try:
                iolock.acquire()
                fw = FileWriter(outfile_path)
                fw.write(rendered_text)
                iolock.release()
            except Exception as e:
                outputlock.acquire()
                stderr("[!] doxx: Unable to write the file '" + outfile_path + "'. Error: " + str(e), exit=0)
                outputlock.release()
                iolock.release()
                sys.exit(1)  # release the iolock before raising SystemExit
                
            outputlock.acquire()
            stdout("[+] doxx: -- " + outfile_path + " ... check")
            outputlock.release()
        

    def project_archive_run(self, key):
        try:
            project_path = key.meta_data['project']
            # Remote .tar.gz or .zip project archives
            if is_url(project_path):
                # pull the remote project and unpack it, define the root directory with the returned value from the function
                root_dir = run_pull(project_path)
                
                # make the path to the 'project.yaml' key file
                project_key_path = make_path(root_dir, 'project.yaml')           
                
                if file_exists(project_key_path):
                    self.write_project_runner_key(project_key_path)  # append the key data to the project.yaml key
                else:
                    stderr("[!] doxx: Unable to locate the 'project.yaml' project settings file in your unpacked archive", exit=1)
                
                # instantiate a new Builder object with the updated 'project.yaml' local file
                builder = Builder(project_key_path)
                builder.run()  # build with the updated 'project.yaml' key
            # Local .tar.gz or .zip project archives
            else:
                project_key_path = self.unpack_and_get_keypath(project_path)
                
                # create an updated 'project.yaml' key file from the remote project template paths and the local user key data
                if file_exists(project_key_path):
                    self.write_project_runner_key(project_key_path)
                else:
                    stderr("[!] doxx: Unable to locate the 'project.yaml' project settings file in your unpacked archive", exit=1)
                    
                # instantiate a new Builder object with the updated 'project.yaml' local file
                builder = Builder(project_key_path)
                builder.run()            # build with the updated 'project.yaml' key
                
                # remove the project archive
                os.remove(project_path)
        except Exception as e:
            stderr("[!] doxx: Unable to build from the project error because of an error.  Error: " + str(e), exit=1)
                
    
    def unpack_and_get_keypath(self, project_path):
        # unpack the archive and get the root directory from the archive
        root_directory = unpack_run(project_path)
        
        if root_directory is None or root_directory == "":
            key_path = None
            try:
                for root, dirs, files in os.path.walk(cwd()):
                    for test_file in files:
                        if test_file == "project.yaml":
                            key_path = make_path(root, dirs, test_file)
            except Exception as e:
                stderr("[!] doxx: Unable to locate the 'project.yaml' project settings file in your unpacked archive. Error: " + str(e), exit=1)
                
            if key_path is None:  # can't find key path
                stderr("[!] doxx: Unable to locate the 'project.yaml' project settings file in your unpacked archive.", exit=1)
            else:
                return key_path  # return key path to the calling method
        else:
            if root_directory == ".":
                key_path = 'project.yaml'  # in current working directory
            else:
                # make the path to the key file
                root_directory = make_path(cwd(), root_directory)
                key_path = make_path(root_directory, 'project.yaml')
            # return the key path to the calling method
            return key_path
        
    
    def write_project_runner_key(self, project_key_path):
        key_data_string = "\n\n"  # maintain a couple of newlines between the build specification section and the key data section
        if len(self.key_data) > 0:
            for the_key in self.key_data:
                key_data_string = key_data_string + the_key + ": " + self.key_data[the_key] + "\n" # recreate the YAML from the local key.yaml file to append to the project.yaml file
        # if there is no key data, not necessary to write any additional data to the file.  when doxx key is instantiated it will handle the lack of replacement data
        
        # append the local key file ('key.yaml') key data to the project meta data to prepare for the build
        try:
            fw = FileWriter(project_key_path)
            fw.append(key_data_string)  # append local key data to the project.yaml file
        except Exception as e:
            stderr("[!] doxx: Unable to write the temporary key file for your project build. Error: " + str(e), exit=1)
        
        
    


#!/usr/bin/env python
# encoding: utf-8

import tarfile
import zipfile
from os import remove
from Naked.toolshed.system import stderr, file_exists


def unpack_run(file_path):
    if tarfile.is_tarfile(file_path):
        return unpack_targz_archive_file(file_path)  # returns the root directory from the unpack function below
    elif zipfile.is_zipfile(file_path):
        return unpack_zip_archive_file(file_path)    # returns the root directory from the unpack function below
    else:
        stderr("[!] doxx: '" + file_path + "' does not appear to be a supported project archive type.  Please review the project archive documentation and try again.", exit=1)


def unpack_targz_archive_file(targz_file_path):
    try:
        tar = tarfile.open(targz_file_path, "r:*")
        # define root directory in the tar archive to return to calling function
        x = 0
        root_dir = None
        for tarinfo in tar:
            if x == 0:
                root_dir = tarinfo.name
                x += 1
            else:
                break
        # unpack the archive
        tar.extractall()
        tar.close()
        # return the root directory to calling function
        return root_dir
    except Exception as e:
        stderr("[!] doxx: Unable to unpack the file '" + targz_file_path + "'. Error: " + str(e))

  
def unpack_zip_archive_file(zip_file_path):
    try:
        zip_archive = zipfile.ZipFile(zip_file_path, 'r')
        # define root directory in the tar archive to return to calling function
        x = 0
        root_dir = None
        for zipinfo in zip_archive.infolist():
            if x == 0:
                root_dir = zipinfo.filename
                x += 1
            else:
                break
        # extract the zip archive
        zip_archive.extractall()
        zip_archive.close()
        
        # if root_dir is actually a file name, then assume root dir is the CWD
        if '.' in root_dir:
            root_dir = "."
        # return the root directory in the zip archive
        return root_dir
    except Exception as e:
        stderr("[!] doxx: Unable to unpack the file '" + zip_file_path + "'. Error: " + str(e))

        
def remove_compressed_archive_file(targz_file_path):
    if file_exists(targz_file_path):
        remove(targz_file_path)
    else:
        pass  # could not find the compressed archive file, raise no error message and abort attempt
    


#!/usr/bin/env python
# encoding: utf-8

import heapq

from doxx.datatypes.package import OfficialPackage
from doxx.utilities.fuzzysearch import FuzzySearcher
from doxx.datatypes.cache import DoxxCache
from Naked.toolshed.system import stdout, stderr
from Naked.toolshed.network import HTTP


def run_search(search_string):
    # check for local cached list of the repositories, if not present pull the remote repository list from Amazon S3 store
    stdout("[*] doxx: Searching remote doxx repositories...")
    master_text = _get_master_text()
    master_list = _get_master_list(master_text)
    search_word_count = len(search_string.split(" "))
    
    # fuzzy search for user search string
    fuzzy = FuzzySearcher(search_string)
    
    # maxheaps for best results and possible results
    best_results = []
    possible_results = []
    
    # indices for maxheap pushes (allows order of matching priorities by FIFO)
    best_index = 0
    possible_index = 0
    
    # iterate through the repository names
    for repository in master_list:
        best_ratio = 0
        continue_match_attempts = True
        
        # full string match attempt
        match_ratio_fullstring = fuzzy.full_string_ratio(repository)
        if match_ratio_fullstring > best_ratio:
            best_ratio = match_ratio_fullstring
        if match_ratio_fullstring == 1.0:    # perfect match, add to good_results
            continue_match_attempts = False
            heapq.heappush(best_results, (-1.0, best_index, repository))  # give highest priority level (1.0) for this match
            best_index += 1  # bump the index number
            
        # perfect match on first token in the repository name
        if continue_match_attempts is True:
            match_ratio_firsttoken = fuzzy.partial_firstindexitem_dashsplit_ratio(repository)
            if match_ratio_firsttoken > best_ratio:
                best_ratio = match_ratio_firsttoken
            if match_ratio_firsttoken == 1.0:
                continue_match_attempts = False
                heapq.heappush(best_results, (-0.98, best_index, repository))  # give priority of 0.98 for perfect match
                best_index += 1  # bump index number
                
        # slice the repository name by same number of characters if length > length of the search string
        if continue_match_attempts is True:
            match_ratio_startslice = fuzzy.partial_startslice_ratio(repository)
            if match_ratio_startslice > best_ratio:
                best_ratio = match_ratio_startslice
            if match_ratio_startslice == 1.0:
                continue_match_attempts = False
                heapq.heappush(best_results, (-0.99, best_index, repository))
                best_index += 1
                
        # (single word search strings ONLY) attempt match for repository tokens split on '-'
        if continue_match_attempts is True and search_word_count == 1:
            match_ratio_dashtokens = fuzzy.partial_dashsplit_tokens_ratio(repository)
            if match_ratio_dashtokens > best_ratio:
                best_ratio = match_ratio_dashtokens
            if match_ratio_dashtokens == 1.0:
                continue_match_attempts = False
                heapq.heappush(best_results, (-0.95, best_index, repository))
                best_index += 1
                
        # (multi word search strings ONLY) attempt match for sequential groups of repository tokens that are same word count as search string word count
        if continue_match_attempts is True and search_word_count > 1:
            match_ratio_nword = fuzzy.partial_nword_ratio(repository)
            if match_ratio_nword > best_ratio:
                best_ratio = match_ratio_nword
            if match_ratio_nword == 1.0:
                continue_match_attempts = False
                heapq.heappush(best_results, (-0.99, best_index, repository))
                best_index += 1
                
        # (multi word search strings ONLY) set intersection + remainder matching attempt
        if continue_match_attempts is True and search_word_count > 1:
            match_ratio_set = fuzzy.partial_set_ratio(repository)
            if match_ratio_set > best_ratio:
                best_ratio = match_ratio_set
            if match_ratio_set == 1.0:
                continue_match_attempts = False
                heapq.heappush(best_results, (-0.97, best_index, repository))
                best_index += 1
                
        # match attempts are complete, determine the quality of match if not previously determined and push to appropriate maxheap to store it
        if continue_match_attempts is True:
            if best_ratio > 0.8:
                heapq.heappush(best_results, (-best_ratio, best_index, repository))
                best_index += 1
            elif best_ratio > 0.6 and best_ratio <= 0.8:
                heapq.heappush(possible_results, (-best_ratio, possible_index, repository))
                possible_index += 1
                
    # report results of the fuzzy search for the user's search term
    final_best_results = _get_maxheap_results_list(best_results)
    final_possible_results = _get_maxheap_results_list(possible_results)
    
    if len(final_best_results) > 0:
        stdout(" ")
        for result in final_best_results:
            stdout(result)
    elif len(final_possible_results) > 0:
        stdout("[*] doxx: There were no good matches for your search.")
        stdout("[*] doxx: Do any of these work? :")
        stdout(" ")
        for result in final_possible_results:
            stdout(result)
        pass  # handle with possible results
    else:
        stdout("[*] doxx: No matches found in the Package Repository.")  ##TODO: add prompt to develop a new package
        
    
def _get_master_text():
    ## check the cache for a cached version of the file with appropriate cache duration
    cache = DoxxCache()
    max_cache_seconds = 86400  # 24 hour cache of the repository list
    if cache.cached_file_exists(cache.package_repo_list_file):
        if cache.does_cache_file_require_update(cache.package_repo_list_file, max_cache_seconds):
            master_list = _pull_official_repository_list()  # pull the master list text from remote
            cache.cache_packagerepo_list(master_list)  # cache it
        else:
            master_list = cache.get_cached_packagerepo_list()  # if cache duration hasn't expired, read the local cached file
    else:
        master_list = _pull_official_repository_list()  # pull the master list text from remote repository if there is no cached file
        cache.cache_packagerepo_list(master_list)  # cache the remote file
    
    # return the master list to the calling function
    return master_list
    

def _pull_official_repository_list():
    package = OfficialPackage()
    master_package_list_url = package.get_master_package_list_url()
    http = HTTP(master_package_list_url)
    try:
        if http.get_status_ok():
            master_list = http.res.text
            return master_list.strip()   # strip additional spaces, blank end lines off of the list
        else:
            stderr("[!] Unable to pull the remote repository list (HTTP status code: " + str(http.res.status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull the remote repository list. Error: " + str(e), exit=1)


def _get_master_list(master_text):
    return master_text.split('\n')


def _get_maxheap_results_list(maxheap_object):
    results_list = []
    for x in range(len(maxheap_object)):
        results_list.append(heapq.heappop(maxheap_object)[-1])  # pop each result off the heap and append to the list
    
    return results_list



#!/usr/bin/env python
# encoding: utf-8

import os
import shutil
import tarfile
import zipfile
from Naked.toolshed.system import stderr


def tar_gzip_package_directory(archive_name, root_dir):
    try:
        current_dir = os.getcwd()  
        archive_gz_name = archive_name + ".tar.gz"
        tar = tarfile.open(archive_gz_name, mode="w:gz", compresslevel=9)    # file writes to current working directory
        os.chdir(root_dir)  # navigate to the root directory to add the files to the archive
        tar.add(".")     # make tar.gz archive
        tar.close()
        os.chdir(current_dir)  # navigate back to user's current working directory
    except Exception as e:
        os.chdir(current_dir)  # if exception was raised, make sure that user is back in their current working directory before raising system exit
        tar.close()
        stderr("[!] doxx: Unable to pack the directory '" + root_dir + "'. Error: " + str(e))
        

def zip_package_directory(archive_name, path):
    try:
        current_dir = os.getcwd()
        archive_name = archive_name + '.zip'
        archive_file_list = []
        os.chdir(path)
        for root, dirs, files in os.walk(os.getcwd()):
            for the_file in files:      
                archive_file_list.append((os.path.relpath(os.path.join(root, the_file))))
        zipper = zipfile.ZipFile(archive_name, 'w')
        for zip_file in archive_file_list:
            zipper.write(zip_file)
        zipper.close()
        shutil.move(archive_name, os.path.join(current_dir, archive_name))
        os.chdir(current_dir)
    except Exception as e:
        os.chdir(current_dir)
        zipper.close()
        stderr("[!] doxx: Unable to pack the directory '" + path + "'. Error: " + str(e))



#!/usr/bin/env python
# encoding: utf-8

import os
import unicodedata
from Naked.toolshed.file import FileReader
from Naked.toolshed.system import directory, make_path
from Naked.toolshed.system import file_exists, stderr
from Naked.toolshed.python import is_py2
from yaml import load_all
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


class DoxxKey(object):
    def __init__(self, inpath):
        # instance variables
        self.meta_data = {}      # holds key meta data (template or templates keys)
        self.key_data = {}       # holds key data (user specified keys)
        self.key_path = inpath
        self.no_replacements = False     # changed to True in _read_yaml if there are no replacement string keys included
        self.multi_template_key = False  # changed to True in the _generate_dir_paths method if 'templates' detected in meta data
        self.project_key = False  # changed to True in the _generate_dir_paths method if 'project' detected in meta data
        
        # define instance variables on object instantiation
        self._read_yaml(inpath)  # define self.meta_data & self.key_data with the yaml key file
        self._cast_values_to_unicode()  # cast non-string values to strings (necessary for the Ink Renderer class)
        self._generate_dir_path(inpath)  # join the directory path to the template files specified in the key (for keys executed from outside of containing directory)
        
        # confirm the integrity of the key file
        self._parse_yaml_for_errors()
    
    def _read_yaml(self, inpath):
        """local YAML doxx key file reader (private method)"""
        if file_exists(inpath):
            fr = FileReader(inpath)
            the_yaml = fr.read()  # FileReader reads in utf-8 encoded bytes data
            the_data = load_all(the_yaml, Loader=Loader)
            i = 0
            for x in the_data:
                if i == 0:
                    # first section of the YAML is meta data
                    self.meta_data = x
                elif i == 1:
                    # second section of the YAML includes the key data
                    self.key_data = x
                else:
                    pass  # ignore any other sections that are included
        
                i += 1  # used to iterate through the_data
            # define the no replacements attribute as True if there is no key data
            if self.key_data is None:
                self.no_replacements = True
        else:
            stderr("[!] doxx: Unable to load the requested key " + inpath + ". Please check the path and try again.", exit=1)
            
                    
    def _cast_values_to_unicode(self):
        unicode_key_data_dict = {}  # new dictionary that will contain the UTF-8 encoded unicode keys and values from self.key_data
        if self.key_data is None or len(self.key_data) == 0:
            key_list = []
        else:
            key_list = self.key_data.keys()
            
        for key in key_list:
            unicode_key = unicodedata.normalize('NFKD', self._create_python_dependent_unicode(key))                    # encode the key
            if self.key_data[key] == None:
                unicode_value = unicodedata.normalize('NFKD', self._create_python_dependent_unicode(""))              # if value == None (i.e. key present but no definition included), replace with empty string = empty string replacement
            else:
                unicode_value = unicodedata.normalize('NFKD', self._create_python_dependent_unicode(self.key_data[key]))  # otherwise normalize it
            unicode_key_data_dict[unicode_key] = unicode_value       # assign the encoded values to the new dictionary
        
        self.key_data = unicode_key_data_dict     # define the instance key_data with the new unicode encoded keys and values
                
    
    def _create_python_dependent_unicode(self, unknown_encoding_string):
        if is_py2():  # python 2 only 
            if isinstance(unknown_encoding_string, unicode):    # test for Python 2 unicode type
                return unknown_encoding_string                  # it is already unicode, just return it
            else:
                return unicode(unknown_encoding_string)         # otherwise, cast to unicode
        else:  # python 3 only
            if isinstance(unknown_encoding_string, str):
                return unknown_encoding_string                  # return unmodified string, py3 strings are unicode by default
            else:
                return str(unknown_encoding_string)             # convert to utf-8 encoded string   
    
    
    def _generate_dir_path(self, inpath):
        """joins the directory path from the current working directory to the template file paths specified in the doxx key meta data (private method)"""
        if self.meta_data is None:
            meta_keys = []
        else:
            meta_keys = self.meta_data.keys()
        
        if 'template' in meta_keys and self.meta_data['template'] is not None:     # single template file request
            pre_file_path = self.meta_data['template']
            if len(pre_file_path) > 6 and (pre_file_path[0:7] == "http://" or pre_file_path[0:8] == "https://"):  # not necessary to build new path if it is a URL
                pass
            else:
                dir_path = directory(inpath)
                # create OS specific path string from the user entered value in 'template:' field
                os_specific_filepath = self.normalize_filepath(pre_file_path)
                self.meta_data['template'] = make_path(dir_path, os_specific_filepath)
        elif 'templates' in meta_keys:  # multi-template file request
            self.multi_template_key = True  # used to detect whether there is a need to process multiple templates with this key
            i = 0  # used for the templates list key during the new path join in following block
            for template in self.meta_data['templates']:  # iterate through template files and join the directory path to the specified template path
                if template == '':
                    pass  # do nothing if it is an empty string, check will be performed in _parse_yaml_for_errors method below.
                else:
                    pre_file_path = template
                    if len(pre_file_path) > 6 and (pre_file_path[0:7] == "http://" or pre_file_path[0:8] == "https://"):  # not necessary to build new path if it is a URL
                        pass
                    else:
                        dir_path = directory(inpath)
                        # create OS specific path string from the user entered values in the 'templates:' field
                        os_specific_filepath = self.normalize_filepath(pre_file_path)
                        self.meta_data['templates'][i] = make_path(dir_path, os_specific_filepath)
                i += 1
        elif 'project' in meta_keys and not self.meta_data['project'] == None:
            self.project_key = True  # indicate that this is a key that uses a project file
            pre_file_path = self.meta_data['project']
            if len(pre_file_path) > 6 and (pre_file_path[0:7] == "http://" or pre_file_path[0:8] == "https://"):  # not necessary to build new path if it is a URL
                pass
            else:
                dir_path = directory(inpath)
                # create OS specific path string from the user entered value in the 'project:' field
                os_specific_filepath = self.normalize_filepath(pre_file_path)
                self.meta_data['project'] = make_path(dir_path, os_specific_filepath)
        else:
            pass  # if the meta data is missing, the check will be performed in the _parse_yaml_for_errors method below. do nothing here
    
    
        
    def normalize_filepath(self, pre_filepath):
        """returns a filepath with the correct OS-dependent path separators"""
        return os.path.normpath(pre_filepath)    
    
    
    def _parse_yaml_for_errors(self):
        # confirm that there are key data
        # confirm that the meta data are complete
        test_meta_data = self.meta_data
        if test_meta_data is None:
            test_metadata_keys = []
        else:
            test_metadata_keys = self.meta_data.keys()
            
        # missing metadata
        if len(test_metadata_keys) == 0:
            stderr("[!] doxx: The metadata header is missing from your key file.  Please include your template or project file path(s).", exit=1)
        
        # meta data does not contain a template, templates, or projects field test
        if 'template' not in test_metadata_keys:
            if 'templates' not in test_metadata_keys:
                if 'project' not in test_metadata_keys:
                    stderr("[!] doxx: There are no template or project files specified in your key. Please complete the meta data section of your key file.", exit=1)
            
        # TOO MANY FIELDS
        # meta data contains both template and templates fields test
        if 'template' in test_metadata_keys and 'templates' in test_metadata_keys:
            stderr("[!] doxx: The 'template' and 'templates' fields are both included in your key file.  Please remove one field and run your command again.", exit=1)
            
        # meta data contains both project and template fields test
        if 'project' in test_metadata_keys and 'template' in test_metadata_keys:
            stderr("[!] doxx: The 'template' and 'project' fields are both included in your key file.  Please remove one field and run your command again.", exit=1)
            
        # meta data contains both project and templates fields test
        if 'project' in test_metadata_keys and 'templates' in test_metadata_keys:
            stderr("[!] doxx: The 'templates' and 'project' fields are both included in your key file.  Please remove one field and run your command again.", exit=1)
            
        # UNDEFINED TEMPLATES
        # meta data template field does not include a template path value
        if 'template' in test_metadata_keys and test_meta_data['template'] == None:
            stderr("[!] doxx: The template field in your key does not include a template path. Please include a path to your template file.", exit=1)
            
        # meta data templates field does not include template paths
        if 'templates' in test_metadata_keys and (test_meta_data['templates'] == None or len(test_meta_data['templates']) == 0):
            stderr("[!] doxx: The templates field in your key does not include template paths.  Please include paths to your template files.", exit=1)
        
        # meta data project field does not include a template path value
        if 'project' in test_metadata_keys and test_meta_data['project'] == None:
            stderr("[!] doxx: The project field in your key is empty. Please include a path to your template file.", exit=1)     
            
        # meta data templates field includes an empty string file path
        if 'templates' in test_metadata_keys:
            for template in test_meta_data['templates']:
                if template == '':
                    stderr("[!] doxx: The templates field in your key file contains an empty file path.  Please fix this and try again.", exit=1)
                    
                    
                    


#!/usr/bin/env python
# encoding: utf-8

import os
from Naked.toolshed.file import FileWriter
from Naked.toolshed.system import stdout, stderr, file_exists, dir_exists

###############
# KEY.YAML STUB - changes to this stub require changes to the datatypes.key._parse_yaml_for_errors method
###############

key_stub = """
---

# Key file documentation: http://doxx.org/usage/keys/

template: ***.doxt
templates: [***.doxt, 'http://***/***.doxt']
project: ***.tar.gz

---

# Enter the key for the template replacement strings below (YAML syntax):
"""


###################
# PROJECT.YAML STUB - changes to this stub require changes to the datatypes.key._parse_yaml_for_errors method
###################

project_yaml_stub = """
---

# Project archive documentation: http://doxx.org/usage/archives/

template: ***.doxt
templates: [***.doxt, 'http://***/***.doxt']

---
"""

##################
# TEMPLATE STUB
##################

template_stub = """
---doxx---

# Template file documentation: http://doxx.org/usage/templates/

# (REQUIRED) enter the file extension for file rendered from this template, leave blank for no extension
#  the use of a '.' in the extension definition is optional (e.g. you can use '.txt' or 'txt')
extension:

# (OPTIONAL) enter a base file name for files generated from this template:
#  default: absent or undefined = use the template base file name for the generated file
basename:

# (OPTIONAL) enter a sub-directory path relative to the KEY FILE for the template file build
#  default : absent or undefined = write to root directory where the user's key file is located
destination-directory:

# (OPTIONAL): change to 'true' for verbatim template writes (i.e. no text replacements)
#  default: false or absent = attempt text replacements in this template file
verbatim: false

---doxx---

"""

### END TEMPLATE STUB


class Maker(object):
    """doxx key, template, and project stub generator class"""
    def __init__(self):
        pass
    
    def make_key(self, outpath):
        try:
            fw = FileWriter(outpath)
            fw.write(key_stub)
            if file_exists(outpath):
                stdout("[+] doxx: The key stub '" + outpath + "' is available in the current directory.")
        except Exception as e:
            stderr("[!] doxx: Unable to write the key stub to disk.  Error: " + str(e), exit=1)
        
    def make_template(self, outpath):
        try:
            fw = FileWriter(outpath)
            fw.write(template_stub)
            if file_exists(outpath):
                stdout("[+] doxx: The template stub '" + outpath + "' is available in the current directory.")
        except Exception as e:
            stderr("[!] doxx: Unable to write the template stub to disk. Error: " + str(e), exit=1)
            
    def make_project(self):
        try:
            # project.yaml file write
            fw_projyaml = FileWriter('project.yaml')
            fw_projyaml.write(project_yaml_stub)
            
            # pkey.yaml file write
            fw_pkey = FileWriter('pkey.yaml')
            fw_pkey.write(key_stub)
            
            # templates directory write
            if not dir_exists('templates'):
                os.mkdir('templates')
                
            # template.doxt file in templates directory
            template_stub_path = os.path.join('templates', 'stub.doxt')
            fw_template = FileWriter(template_stub_path)
            fw_template.write(template_stub)
            
            # confirm for user
            if file_exists('project.yaml'):
                stdout("[+] doxx: 'project.yaml' ... check")
            else:
                stderr("[!] doxx: There was an error writing the 'project.yaml' key file to your project directory")
                
            if file_exists('pkey.yaml'):
                stdout("[+] doxx: 'pkey.yaml' ... check")
            else:
                stderr("[!] doxx: There was an error writing the 'pkey.yaml' key file to your project directory")
                
            if file_exists(template_stub_path):
                stdout("[+] doxx: '" + template_stub_path + "' ... check")
            else:
                stderr("[!] doxx: There was an error writing the '" + template_stub_path + "' template file to your project directory")
        except Exception as e:
            stderr("[!] doxx: Unable to write project files to disk.  Error: " + str(e), exit=1)
            
        



#!/usr/bin/env python
# encoding: utf-8

from doxx.datatypes.cache import DoxxCache
from doxx.datatypes.package import OfficialPackage
from Naked.toolshed.system import stdout, stderr
from Naked.toolshed.network import HTTP


def run_repoupdate():
    cache = DoxxCache()
    
    # confirm that user is not attempting updates too rapidly (10 sec lockout)
    lockout_time = 10
    # test list.txt file
    if cache.cached_file_exists(cache.package_repo_list_file):
        if not cache.does_cache_file_require_update(cache.package_repo_list_file, lockout_time):
            stderr("[!] doxx: There is a " + str(lockout_time) + " second lockout on repository file updates. Take a break, then try again.", exit=1)
            
    # test packages.json file
    if cache.cached_file_exists(cache.package_repo_json_file):
        if not cache.does_cache_file_require_update(cache.package_repo_json_file, lockout_time):
            stderr("[!] doxx: There is a " + str(lockout_time) + " second lockout on repository file updates. Take a break, then try again.", exit=1)
    
    stdout("[*] doxx: Pulling the list of packages in the Package Repository")
    # pull the master list of the packages included in the Package Repository
    master_list = _pull_official_repository_list()
    if len(master_list) > 0:
        cache.cache_packagerepo_list(master_list)
    else:
        stderr("[!] doxx: Unable to update the " + cache.package_repo_list_file + " file.")
        
    if not cache.cached_file_exists(cache.package_repo_list_file):
        stderr("[!] doxx: Unable to update the " + cache.package_repo_list_file + " file.")
        
    # pull the descriptions of the packages included in the Package Repository
    stdout("[*] doxx: Pulling the descriptions of packages in the Package Repository")
    master_description_json = _pull_official_repository_json()
    if len(master_description_json) > 0:
        cache.cache_packagerepo_json(master_description_json)
    else:
        stderr("[!] doxx: Unable to update the " + cache.package_repo_json_file + " file.", exit=1)
        
    if not cache.cached_file_exists(cache.package_repo_json_file):
        stderr("[!] doxx: Unable to update the " + cache.package_repo_json_file + " file.", exit=1)
    
    stdout("[*] doxx: repoupdate complete")


def _pull_official_repository_list():
    package = OfficialPackage()
    master_package_list_url = package.get_master_package_list_url()
    http = HTTP(master_package_list_url)
    try:
        if http.get_status_ok():
            master_list = http.res.text
            return master_list.strip()   # strip additional spaces, blank end lines off of the list
        else:
            stderr("[!] Unable to pull the remote repository list (HTTP status code: " + str(http.res.status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull the remote repository list. Error: " + str(e), exit=1)


def _pull_official_repository_json():
    package = OfficialPackage()
    master_package_json_url = package.get_master_package_description_json_url()
    http = HTTP(master_package_json_url)
    try:
        if http.get_status_ok():
            master_list = http.res.text
            return master_list.strip()   # strip additional spaces, blank end lines off of the list
        else:
            stderr("[!] Unable to pull the remote repository package descriptions (HTTP status code: " + str(http.res.status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull the remote repository package descriptions. Error: " + str(e), exit=1)  


#!/usr/bin/env python
# encoding: utf-8

import gzip
from os import remove, rename
from Naked.toolshed.network import HTTP
from Naked.toolshed.file import FileWriter
from Naked.toolshed.system import stderr, stdout, file_exists
from Naked.toolshed.python import is_py3
from doxx.commands.unpack import unpack_run


def run_pull(url):
    # URL pulls for project archive packages, gzip files, text files
    if is_url(url):
        file_name = get_file_name(url)
        if len(file_name) == 0:
            file_name = "pullfile"
        
        # begin file pull
        stdout("[*] doxx: Pulling file...")
            
        if is_tar_gz_archive(file_name):
            root_dir = None
            try:
                pull_binary_file(url, file_name)      # pull remote file
            except Exception as e:
                stderr("[!] doxx: Unable to pull the tar.gz project. Error: " + str(e), exit=1)
            stdout("[*] doxx: Unpacking...")
            try:
                root_dir = unpack_archive(file_name)             # unpack archive and define the root directory
            except Exception as e:
                stderr("[!] doxx: Unable to unpack the compressed project file. Error: " + str(e), exit=1)
            if file_exists(file_name):
                remove_file(file_name)                # remove the archive file
            if file_exists('pkey.yaml'):
                if not file_exists('key.yaml'):
                    rename('pkey.yaml', 'key.yaml')   # change name of pkey.yaml to key.yaml if there is not already a key.yaml file
            return root_dir                           # return the root directory path for calling code that needs it
        elif is_zip_archive(file_name):
            root_dir = None
            try:
                pull_binary_file(url, file_name)      # pull remote file
            except Exception as e:
                stderr("[!] doxx: Unable to pull the .zip project. Error: " + str(e), exit=1)
            stdout("[*] doxx: Unpacking...")
            try:
                root_dir = unpack_archive(file_name)             # unpack archive and define the root directory
            except Exception as e:
                stderr("[!] doxx: Unable to unpack the compressed project file. Error: " + str(e), exit=1)
            if file_exists(file_name):
                remove_file(file_name)                # remove the arhcive file
            if file_exists('pkey.yaml'):
                if not file_exists('key.yaml'):
                    rename('pkey.yaml', 'key.yaml')   # change name of pkey.yaml to key.yaml if there is not already a key.yaml file      
            return root_dir                           # return the root directory path for calling code that needs it
        elif is_gzip_file(file_name):
            try:
                pull_binary_file(url, file_name)      # pull the remote gzip file
            except Exception as e:
                stderr("[!] doxx: Unable to pull the compressed file. Error: " + str(e), exit=1)
            stdout("[!] doxx: Decompressing...")
            try:
                decompress_gzip(file_name)           # decompress the file text
            except Exception as e:
                stderr("[!] doxx: Unable to decompress the gzip file. Error: " + str(e), exit=1)
            if file_exists(file_name):
                remove_file(file_name)               # remove the gzip compressed file and leave the decompressed text file
        else:
            try:
                pull_text_file(url, file_name)       # it is assumed to be a plain text template or key file, pull the text
            except Exception as e:
                stderr("[!] doxx: Unable to pull the requested file. Error: " + str(e), exit=1)
    else:
        # SHORT CODES for Github repository, CDNJS, etc
        if "/" in url:
            short_code = url
            short_code_parts = short_code.split('/')
            
            if short_code.startswith('cdnjs:'):
                pass  # add code for cdnjs pulls
            else:
                # default to Github repositories
                if len(short_code_parts) == 2:
                    if "#" in short_code_parts[1]:
                        # contains a request for a specific branch of the Github repository
                        user = short_code_parts[0]
                        if "#" in user:
                            stderr("[!] doxx: the short code for Github repositories is not properly formed")
                            stderr("[!] doxx: the syntax is [user]/[repository]#[branch]", exit=1)
                        repo_parts = short_code_parts[1].split('#')
                        repo = repo_parts[0]
                        branch = repo_parts[1]
                        targz_filename = repo + "-" + branch + ".tar.gz"
                        url = "https://github.com/{{user}}/{{repository}}/archive/{{branch}}.tar.gz"
                        url = url.replace("{{user}}", user)
                        url = url.replace("{{repository}}", repo)
                        url = url.replace("{{branch}}", branch)
                        user_message = "[*] doxx: Pulling branch '" + branch + "' of Github repository '" + user + "/" + repo + "'..."
                    else:
                        # default to the master Github repository branch
                        user = short_code_parts[0]
                        if "#" in user:
                            stderr("[!] doxx: the short code for Github repositories is not properly formed")
                            stderr("[!] doxx: the syntax is [user]/[repository]#[branch]", exit=1)                        
                        repo = short_code_parts[1]
                        targz_filename = repo + "-master.tar.gz"
                        url = "https://github.com/{{user}}/{{repository}}/archive/master.tar.gz"
                        url = url.replace("{{user}}", user)
                        url = url.replace("{{repository}}", repo)
                        user_message = "[*] doxx: Pulling master branch of Github repository '" + user + "/" + repo + "'..."
                     
                    # notify user of the pull   
                    stdout(user_message)
                        
                    try:
                        pull_binary_file(url, targz_filename)  # pull the archive file
                    except Exception as e:
                        stderr("[!] doxx: Unable to pull the Github repository.  Error: " + str(e), exit=1)
                        
                    if file_exists(targz_filename):
                        try:
                            unpack_archive(targz_filename)  # unpack the archive locally
                            remove(targz_filename)          # remove the archive file
                        except Exception as e:
                            stderr("[!] doxx: Unable to unpack the pulled Github repository. Error: " + str(e), exit=1)
                    else:
                        stderr("[!] doxx: The Github repository pull did not complete successfully.  Please try again.")
                else:
                    stderr("[!] doxx: short code syntax for Github repository pulls:", exit=0)
                    stderr("  $ doxx pull [user]/[repository]")
                    stderr("with an optional branch")
                    stderr("  $ [user]/[repository]#[branch]")
            
        # PROJECT PACKAGES - official repository package pulls
        else:  
            from doxx.datatypes.package import OfficialPackage
            package_name = url
            package = OfficialPackage()
            package_url = package.get_package_targz_url(package_name)
            package_archive_name = package_name + ".tar.gz"
            root_dir = None
            # pull the package archive file
            stdout("[*] doxx: Pulling package '" + package_name + "'...")
            try:
                pull_binary_file(package_url, package_archive_name)
            except Exception as e:
                stderr("[!] doxx: Unable to pull the doxx repository package.  Error: " + str(e), exit=1)
            # unpack the archive file
            stdout("[*] doxx: Unpacking...")
            try:
                root_dir = unpack_archive(package_archive_name)             # unpack archive and define the root directory
            except Exception as e:
                stderr("[!] doxx: Unable to unpack the project package. Error: " + str(e), exit=1)            
            # remove the archive file
            if file_exists(package_archive_name):
                remove_file(package_archive_name)                # remove the archive file
            if file_exists('pkey.yaml'):
                if not file_exists('key.yaml'):
                    rename('pkey.yaml', 'key.yaml')              # change name of pkey.yaml to key.yaml if there is not already a key.yaml file            
            return root_dir                                      # return the root directory path for calling code that needs it            


def is_url(url):
    """test for HTTP and HTTPS protocol in the string"""
    if len(url) > 6 and (url[0:7] == "http://" or url[0:8] == "https://"):
        return True
    else:
        return False


def is_tar_gz_archive(file_name):
    """test for tar.gz file archive"""
    if file_name.lower().endswith('.tar.gz') or file_name.lower().endswith('.tar.gzip'):
        return True
    else:
        return False


def is_zip_archive(file_name):
    """test for zip file archive extension"""
    if file_name.lower().endswith('.zip'):
        return True
    else:
        return False


# must be tested AFTER tests for tar.gz archive because both will test true
def is_gzip_file(file_name):
    """test for gzip compressed file extension"""
    if file_name.lower().endswith('.gz') or file_name.lower().endswith('.gzip'):
        return True
    else:
        return False


def get_file_name(url):
    """returns the filename from a URL"""
    split_url = url.split('/')
    return split_url[-1]


def pull_binary_file(url, binary_file_name):
    """pulls a remote binary file and writes to disk"""
    # pull the binary file data
    http = HTTP(url)
    try:
        if http.get_status_ok():
            binary_data = http.res.content    
            # write binary data to disk
            try:
                fw = FileWriter(binary_file_name)
                fw.write_bin(binary_data)
            except Exception as e:
                stderr("[!] doxx: File write failed for '" + binary_file_name + "'.  Error: " + str(e), exit=1)
        else:
            fail_status_code = http.res.status_code
            if fail_status_code == 404:
                stderr("[!] doxx: Unable to pull the file because it does not appear to exist. (HTTP status code: " + str(fail_status_code) + ")", exit=1)
            else:
                stderr("[!] doxx: Unable to pull '" + url + "'. (HTTP status code: " + str(fail_status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull '" + url + "'. Error: " + str(e), exit=1)


def pull_text_file(url, text_file_name):
    """pulls a remote text file and writes to disk"""
    # pull the binary file data
    http = HTTP(url)
    try:
        if http.get_status_ok():
            text_data = http.res.text
            # write text data to disk
            try:
                fw = FileWriter(text_file_name)
                fw.write(text_data)
            except Exception as e:
                stderr("[!] doxx: File write failed for '" + text_file_name + "'.  Error: " + str(e), exit=1)
        else:
            fail_status_code = http.res.status_code
            stderr("[!] doxx: Unable to pull '" + url + "' (HTTP status code " + str(fail_status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull '" + url + "'. Error: " + str(e), exit=1)


def unpack_archive(archive_file_name):
    """unpacks a tar.gz or zip file archive and writes to local disk"""
    root_dir = unpack_run(archive_file_name)  # root directory of unpacked archive returned from the unpack function, returned from this function if calling code needs it
    return root_dir


def decompress_gzip(gz_filename):
    """decompress gzip compressed file"""
    # decompress the gzip'd file
    f = gzip.open(gz_filename, 'rb')
    file_content = f.read()
    f.close()
    
    # monkeypatch for Python 3 byte string read issue
    if is_py3():
        file_content = str(file_content)
    
    # get the base file name for the decompressed file write
    filename_split = gz_filename.split('.')
    if len(filename_split) == 2:
        basename = filename_split[0]
    elif len(filename_split) > 2:
        basename = filename_split[0] + '.' + filename_split[1]  # concatenate first two parts of file name (e.g. example + '.' + 'tar' )
    else:
        basename = gz_filename
        
    # write the file locally
    fw = FileWriter(basename)
    fw.write(file_content)


def remove_file(file_name):
    """removes archive file that was pulled from a remote source"""
    if file_exists(file_name):
        remove(file_name)
        


#!/usr/bin/env python
# encoding: utf-8

from os.path import basename, splitext, normpath
from Naked.toolshed.file import FileReader
from Naked.toolshed.network import HTTP
from Naked.toolshed.system import make_path

from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
    

class DoxxTemplate(object):
    """A doxx template class that maintains state of user designed templates during the rendering process"""
    def __init__(self, inpath):
        self.inpath = inpath
        self.raw_text = ""
        self.text = ""
        self.meta_data = {}     # holds meta data from the header of the file
        self.verbatim = False   # should file write take place verbatim (without text replacements), default False, changed to True by user meta data verbatim field
        self.extension = ""     # stored in the format '.txt'
        self.basename = ""      # base filename for the out write file path
        self.outfile = ""       # write file path for use by calling code
    
    def load_data(self):
        fr = FileReader(self.inpath)
        self.raw_text = fr.read()

    def split_data(self):
        parsed_text = self.raw_text.split("---doxx---")

        if len(parsed_text) == 3:  # should split into three sections (0 = before first ---doxx---, 1 = meta data, 2 = template text data after second ---doxx---)
            self.meta_data = load(parsed_text[1], Loader=Loader)
            self.text = parsed_text[2][1:]  # define self.text with the template data from the file, the [1:] slice removes /n at end of the delimiter        
        else:
            self.meta_data = {}
            self.text = u""
    
    def parse_template_text(self):
        """parses doxx template meta data YAML and main body text and defines instance variables for the DoxxTemplate (public method)"""

        meta_keys = self.meta_data.keys()

        # if user did not enter an extension type, 
        if 'extension' not in meta_keys or self.meta_data['extension'] == None:
            self.extension = ""  # default to no extension if user did not enter the meta data field, or the meta data field has no definition
        else:
            self.extension = self.meta_data['extension']  # define the instance extension attribute with the value from the template meta data
        
        # define rendered file destination directory relative to current working directory
        if 'destination_directory' not in meta_keys or self.meta_data['destination_directory'] == None:
            dest_dir = u""
        else:
            # replace user entered path with OS specific separators and re-define the destination directory attribute
            self.meta_data['destination_directory'] = self.normalize_filepath(self.meta_data['destination_directory'])
            dest_dir = self.meta_data['destination_directory']

        # define rendered file extension as instance attribute
        if len(self.extension) > 0:  # if no extension (user did not enter extension field or did not define an extension), then skip '.' check
            if self.extension[0] == ".":
                pass
            else:
                self.extension = "." + self.extension  # add a period if the user did not include it

        # define rendered file base file name (different approach for local vs. remote paths)
        if 'basename' not in meta_keys or self.meta_data['basename'] == None:
            # if it is a URL, split by '/' and obtain file path from the URL
            if len(self.inpath) > 6 and (self.inpath[0:7] == "http://" or self.inpath[0:8] == "https://"):
                url_path = self.inpath.split("/")
                filename_with_extension = splitext(url_path[-1])  # get the filename from the '/' split url :: url_path[-1] = URL filename, [0] index is the basename from that filename.extension
                self.basename = filename_with_extension[0]
            else:  # local files
                self.basename = splitext(basename(self.inpath))[0]
        else:
            self.basename = self.meta_data['basename']

        file_name = self.basename + self.extension  # local temp file_name variable, modified below to final write path

        # make the outfile path to be used for the rendered file write
        if len(dest_dir) > 0:
            self.outfile = make_path(dest_dir, file_name)
        else:
            self.outfile = file_name
        
        # test for user request to write the file out verbatim, skips the text replacement step during the build
        if 'verbatim' in meta_keys and self.meta_data['verbatim'] == True:
            self.verbatim = True
        else:
            self.meta_data['verbatim'] = False
            # self.verbatim is False by default, define the meta_data key with False value in case user included it without definining it

    def parse_template_for_errors(self):
        """evaluates template file for presence of meta data header and template text.  Returns a two-tuple. tuple[0] = True if error identified, False if error not identified.  tuple[1] = error message string"""
        # confirm meta data contains data
        if self.meta_data is None or len(self.meta_data) == 0:
            error_message = u"[!] doxx: The template file '" + self.inpath + "' is not properly formatted.  Please include the required meta data block between '---doxx---' delimiters at the top of your file."
            return (True, error_message)
        # confirm that there is template text
        elif self.text is None or len(self.text) < 5:  # if self.text not defined or length of the string < 5 chars (because {{x}} == 5 so must not include any replacement tags)
            error_message = u"[!] doxx: Unable to parse template text from the template file '" + self.inpath + "'. Please include a template in order to render this file."
            return (True, error_message)
        else:
            return(False, "no message")
        

    def normalize_filepath(self, pre_filepath):
        """returns a filepath with the correct OS-dependent path separators"""
        return normpath(pre_filepath)  



class RemoteDoxxTemplate(DoxxTemplate):
    def __init__(self, inpath):
        DoxxTemplate.__init__(self, inpath)

    def load_data(self):
        """overloaded load_data method that retrieves data from a remote file using HTTP or HTTPS protocol instead of a local file"""
        http = HTTP(self.inpath)
        
        if http.get_status_ok():
            import unicodedata
            norm_text = unicodedata.normalize('NFKD', http.res.text)  # normalize unicode data to NFKD (like local file reads)
            self.raw_text = norm_text
            return (True, "no message")
        else:
            fail_status_code = http.res.status_code
            error_message = "[!] doxx: Unable to load the remote template file '" + self.inpath + "' (HTTP status code " + str(fail_status_code) + ")"
            return (False, error_message)
            
        
        


#!/usr/bin/env python
# encoding: utf-8

import json
from doxx.datatypes.cache import DoxxCache
from doxx.datatypes.package import OfficialPackage

from Naked.toolshed.network import HTTP
from Naked.toolshed.system import stdout, stderr


def run_whatis(package_name):
    cache = DoxxCache()
    max_cache_seconds = 86400  # 24 hour file cache
    stdout("[*] doxx: Looking up the package description...")
    if cache.cached_file_exists('packages.json'):  # does the cached file already exist
        if cache.does_cache_file_require_update(cache.package_repo_json_file, max_cache_seconds):  # if so, has the cache time elapsed?
            master_descriptions = _pull_official_repository_descriptions()  # pull new file
            cache.cache_packagerepo_json(master_descriptions)               # push it into the file cache
        else:
            master_descriptions = cache.get_cached_packagerepo_json()       # get the cached file if there is no need to pull a new version
    else:
        master_descriptions = _pull_official_repository_descriptions()      # doesn't exist, go get it
        cache.cache_packagerepo_json(master_descriptions)                   # try to cache it
    
    if len(master_descriptions) > 0:  # dictionary from JSON that contains the package descriptions by package name key
        descriptions_dict = json.loads(master_descriptions)
        test_package_name = package_name.lower().strip()
        if test_package_name in descriptions_dict:
            stdout(" ")
            stdout("    Package: " + test_package_name)
            stdout("Description: " + descriptions_dict[test_package_name])
        else:
            stderr("[!] doxx: Unable to locate the package '" + test_package_name + "'", exit=1)
    else:
        stderr("[!] doxx: Unable to read the descriptions file.  It appears to be empty...", exit=1)
        

def _pull_official_repository_descriptions():
    package = OfficialPackage()
    master_package_descriptions = package.get_master_package_description_json_url()
    http = HTTP(master_package_descriptions)
    try:
        if http.get_status_ok():
            master_descriptions = http.res.text
            return master_descriptions.strip()
        else:
            stderr("[!] doxx: Unable to pull remote repository descriptions (HTTP status code: " + str(http.res.status_code) + ")", exit=1)
    except Exception as e:
        stderr("[!] doxx: Unable to pull remote repository descriptions Error: " + str(e) + ")", exit=1)
        